'use strict';
const constants = require(".././constants_local.json")
const secrets = require('./secrets.json')
const config = require('./config.json')
const libSupport = require('./lib')
libSupport.updateConfig()
const node_id = config.id
const {spawn } = require('child_process')
const execute = require('./execute')
const fs = require('fs')
const fetch = require('node-fetch');
const os = require('os');
const dgram = require('dgram');
const server = dgram.createSocket('udp4');
const udpProxy = dgram.createSocket('udp4');

const WINDOW_SIZE = 10

let struct = require('jspack');
struct = struct.jspack

let metadataDB = `http://${secrets.couchdb_username}:${secrets.couchdb_password}@${constants.couchdb_host}`
metadataDB = metadataDB + "/" + constants.db.function_meta + "/"

let usedPort = new Map(); //map of used ports by function
let functionToResourceMap={} // f
let new_starting_function={}
let coldstart_request_queue = []

const kafka = require('kafka-node')
const logger = libSupport.logger

const local_repository = __dirname + "/local_repository/"
const host_url = "http://" + constants.master_address + ":" + constants.master_port

let Producer = kafka.Producer,
    client = new kafka.KafkaClient({ 
        kafkaHost: constants.network.external.kafka_host,
        autoConnect: true
    }),
    producer = new Producer(client),
    Consumer = kafka.Consumer

function getRandomInt(min, max) {
    min = Math.ceil(min);
    max = Math.floor(max);
    return Math.floor(Math.random() * (max - min + 1)) + min;
}
function dispatch_coldstart_request()
{
    //dispatch the coldstart request
    
    // console.log("dispatch the coldstart request !!!")
    // let lookbackWindow = Math.min(WINDOW_SIZE, coldstart_request_queue.length)
    let lookbackWindow = coldstart_request_queue.length

    for (let i = 0; i < lookbackWindow; i++) {
        let req = coldstart_request_queue.shift()
        console.log("coldstart req in queue : ", req)
        // dispatch the request 
        let id = req.request.functionHash + req.request.runtime
        let data = 101
        if(functionToResourceMap[id])
        {
            console.log("function is running so dispatching request : " , functionToResourceMap[id])
            let packet = packPacket({
                chain_id: 0,
                exec_id: 0,
                function_id: functionToResourceMap[id].functionHash.slice(9),
                data,
                function_count: 1,
                // autostart: 1,
                // fno : 255
            })
            let proxy_port = functionToResourceMap[id].portExternal
            console.log("proxing packet to function  : ", packet)

            try {
                udpProxy.send(packet, 0, packet.length, proxy_port, functionToResourceMap[id].node_id, function (err, bytes) {
                    if (err)
                        console.log(err)
                    console.log("UDP coldstart request sent to function")
                })
            } catch (e) {
                console.log("error proxying coldstart request",e)
            }
        }
        else{
            coldstart_request_queue.push(req)
        }
	    // logger.info(req)
        // console.log("dispatch : ",req.body, "params : ",req.params, "uri : ",req.uri)
        // let runtime = req.body.runtime
        // let functionHash = req.params.id
        // if (!db.has(functionHash + runtime)) {
        //     db.set(functionHash + runtime, [])
        //     db.get(functionHash + runtime).push({ req, res })
            
        //     let payload = [{
        //         topic: constants.topics.hscale,
        //         messages: JSON.stringify({ runtime, functionHash })
        //     }]
        //     producer.send(payload, function () { })

        //     speculative_deployment(req, runtime)
        // } else {
        //     logger.info("deployment process already started waiting")
        //     db.get(functionHash + runtime).push({ req, res })
        // }

        
    }

}

libSupport.makeTopic(node_id).then(() => {
    logger.info("node topic created")
    let consumer = new Consumer(client,
        [
            { topic: node_id, partition: 0, offset: 0 },
            { topic: constants.topics.remove_function_intstance},
            { topic: constants.topics.deployed }
        ],
        [
            { autoCommit: true }
        ])
    consumer.on('message', function (message) {
        // logger.info(message);
        let topic = message.topic
        message = message.value
        message = JSON.parse(message)
        console.log("message received on kafka : ",message)
        let runtime = message.runtime
        let functionHash = message.functionHash
        // let resource_id = message.resource_id
        let resource_id = libSupport.makeid(constants.id_size)
        
        // let port = message.port
        let port = libSupport.getPort(usedPort)
        let mac = constants.daemon_mac
        let firststart = false
        let executordata = {}
        /**
         * Download necessary files (function file) and Start resource deployment
         */
        if (message.type === "execute" && topic === node_id) {
            logger.info("Received Deployment request for resource_id: " + resource_id);
            fetch(metadataDB + functionHash).then(res => res.json())
            .then(json => {
                console.log("metadata", json);
                
                libSupport.download(host_url + "/repository/" + functionHash + ".js", local_repository + functionHash + ".js").then(() => {
                    let metadata = {
                        resource_id, functionHash,
                        runtime, port, mac,
                        resources: {
                            memory: json.memory
                        },
                        firststart,
                        executordata
                    }
                    startWorker(local_repository, producer, metadata)
                })
            }).catch(err => {
                logger.error("something went wrong" + err.toString())
            });
            
        }
        else if (topic==constants.topics.remove_function_intstance)
        {
            logger.info("function closed !!! Freeing resources");
            console.log("fc message : ", message)  
            // if(usedPort.has(port))   
            // usedPort.delete(port)
            let id = message.functionHash + message.runtime
            if(functionToResourceMap[id])
                delete functionToResourceMap[id]
        }
        else if (topic == constants.topics.deployed ) {
            // function deployed store metadata in functiontoResourceMap
            // message = JSON.parse(message)
            //fuction is deployed successfully
            console.log("message deployed : ", message)
            let id = message.functionHash + message.runtime
            if(new_starting_function[id])
                delete new_starting_function[id]
           
                
            console.log("\nthere is no function for function ID so creating one\n")
            functionToResourceMap[id] = {}    
            functionToResourceMap[id] = {
                "functionHash": message.functionHash,
                "resource_id": resource_id,
                "function_runtime": message.runtime,
                "node_id": message.node_id, 
                "portExternal": message.portExternal,
                "mac": message.mac,
                "open_request_count": 0,
                "load": 0     
            }
                
            console.log("function to resource MAP : ", functionToResourceMap)
                
        }


    })
})

/**
 * download and start grunt
 */
libSupport.download(constants.grunt_host, "grunt", false).then(() => {
    logger.info("Downloaded grunt binary from repository")
    fs.chmod('grunt', 0o755, (err) => {
        logger.info("grunt made executable. Starting grunt")
        let grunt = spawn('./grunt', [node_id])
        grunt.stdout.on('data', data => {
            // logger.info(data.toString());

        })

        grunt.stderr.on('data', data => {
            // logger.info(data.toString());

        })
        grunt.on('close', (code) => {
            logger.info("Grunt exited with exit code", code);

        })
    })

})

    
/**
 * Start a worker executor of the runtime type
 * @param {String} local_repository 
 * @param {String} functionHash 
 * @param {String} resource_id 
 * @param {String} producer 
 * @param {String} runtime 
 * @param {Number} port 
 */
function startWorker(local_repository, producer, metadata) {
    let runtime = metadata.runtime
    console.log(metadata);
    
    logger.info(`Using port ${metadata.port} for functionHash ${metadata.functionHash}`)
    
    if (runtime === "isolate")
        execute.runIsolate(local_repository, metadata)
        .catch(err => {
            logger.error("=====================deployment failed=========================");
            logger.error(err)
            producer.send([{
                topic: constants.topics.deployed,
                messages: JSON.stringify({
                    "status": false,
                    resource_id: metadata.resource_id,
                    "reason": "isolate exit"
                })
            }], () => { })
        })
    else if (runtime === "process")
//	console.log("rutime is process : ",metadata)
        execute.runProcess(local_repository, metadata)
        .catch(err => {
            logger.error("=====================deployment failed=========================");
            producer.send([{ topic: constants.topics.deployed,
                messages: JSON.stringify({ 
                "status": false, 
                resource_id: metadata.resource_id,
                "reason": "process exit"
            }) }], () => { })
        })
    else if (runtime === "container")
    {
        console.log("rutime is container : ",metadata)
        execute.runContainer(metadata)
    }
    else {
        producer.send(
            [{
                topic: "response2",
                messages: JSON.stringify({ status: "unknown runtime" })
            }], () => { })

        return
    }
    
}

function heartbeat() {
    let info = {
        free_mem: os.freemem(),
        cpu_count: os.cpus().length,
        total_mem: os.totalmem(),
        avg_load: os.loadavg()
    }
    let payload = [{
        topic: constants.topics.heartbeat,
        messages: JSON.stringify({
            "address": node_id,
            "portExternal": constants.daemon_port,
            "mac": constants.daemon_mac,
            "system_info": info,
            "timestamp": Date.now()
        })
    }]
    // console.log("daemon system info : ", payload)
    producer.send(payload, function(cb) {})
}

// TODO 2: implement packer deparser for the udp packet
// TODO 3: create UPD server to get the coldstart request 
server.on('error', (err) => {
    console.log(`server error:\n${err.stack}`);
    server.close();
});

server.on('message', (msg, rinfo) => {
    console.log("message", msg)
    let req = unpackPacket(msg)
    console.log(req, typeof req);
    // get the coldstart request and start the function
    // logger.info("Received Deployment UDP request for resource_id: " + resource_id);
    let functionHash = "function_" + req.function_id
    // let resource_id = 'aaa'
    let resource_id = libSupport.makeid(constants.id_size)
    // let runtime = 'process'
    let runtime = constants.runtime 
    // let port = // create random unused port and assign to function
    let port = libSupport.getPort(usedPort)
    let mac = constants.daemon_mac
    let firststart=true
    let executordata = req.data
    let id = functionHash + runtime

    if(new_starting_function[id])
    {
        // function is already running so no need to start new function instance
        // add the request to the request queue and proxy the request
        console.log("function is already starting in some time. Queueing the request")
        let request = {functionHash, runtime }
        console.log("request : ", request)
        coldstart_request_queue.push({request});
    }
    else{
        // there is no function instance running for fid
        // start the new function
        new_starting_function[id] = 1
        console.log("Starting new function for fid  : ", functionHash)
        logger.info("Received Deployment UPD request")
        fetch(metadataDB + functionHash).then(res => res.json())
            .then(json => {
                console.log("metadata", json);
                
                libSupport.download(host_url + "/repository/" + functionHash + ".js", local_repository + functionHash + ".js").then(() => {
                    let metadata = {
                        resource_id, functionHash,
                        runtime, port, mac,
                        resources: {
                            memory: json.memory
                        }, 
                        firststart,
                        executordata
                    }        
                    startWorker(local_repository, producer, metadata)
            })
            }).catch(err => {
                logger.error("something went wrong" + err.toString())
            });
    }


    // lastRequest = Date.now()
    // console.log("network stack time", lastRequest - payload.t1)
    // totalRequest++
    // executor(msg).then(result => {
    //     result = packPacket(msg)
    //     let port = 10000 + getRandomInt(0, 10)
    //     try {
    //         udpProxy.send(msg, 0, msg.length, port, rinfo.address, function (err, bytes) {
    //             if (err)
    //                 console.log(err)
    //             console.log("response via UDP")
    //         })
    //     } catch (e) {
    //         console.log(e)
    //     }
    // })
});

function unpackPacket(packet) {
    // let buffer = new Array(1024)
    let chain_id = null, exec_id = null, function_count = null, function_id = null, data = null
    let base = 0, f0, f1, f2, f3, f4, t1, t2, t3, t4
    chain_id = struct.Unpack(">I", packet, base)
    base += 4
    exec_id = struct.Unpack(">I", packet, base)
    base += 4
    function_id = struct.Unpack(">I", packet, base)
    base += 4
    data = struct.Unpack(">I", packet, base)
    base += 4
    function_count = struct.Unpack("I", packet, base)
    base += 4

    f0 = struct.Unpack("B", packet, base)
    base += 1
    f1 = struct.Unpack("B", packet, base)
    base += 1
    f2 = struct.Unpack("B", packet, base)
    base += 1
    f3 = struct.Unpack("B", packet, base)
    base += 1
    f4 = struct.Unpack("B", packet, base)
    base += 1

    t1 = struct.Unpack("I", packet, base)
    base += 8
    t2 = struct.Unpack("I", packet, base)
    base += 8
    t3 = struct.Unpack("I", packet, base)
    base += 8
    t4 = struct.Unpack("I", packet, base)

    console.log("chain_id", chain_id, "exec_id", exec_id, "data", data, "function_count", function_count, "function_id", function_id)

    return {
        chain_id: chain_id[0],
        exec_id: exec_id[0],
        data: data[0],
        function_count: function_count[0],
        function_id: function_id[0],
        f0, f1, f2, f3, f4, t1, t2, t3, t4
    }
}

function packPacket(dataPacket) {

    let message = new Array(1024)
    let base = 0, chain_id, exec_id, function_id, data, function_count
    chain_id = struct.PackTo(">I", message, base, [dataPacket.chain_id])
    base += 4
    exec_id = struct.PackTo(">I", message, base, [dataPacket.exec_id])
    base += 4
    function_id = struct.PackTo(">I", message, base, [dataPacket.function_id])
    base += 4
    data = struct.PackTo(">I", message, base, [dataPacket.data])
    base += 4
    function_count = struct.PackTo("B", message, base, [dataPacket.function_count])
    message = Buffer.from(message)
    return message
}

server.on('listening', () => {
    const address = server.address();
    console.log(`server listening ${address.address}:${address.port}`);
});

// server.bind(port, "192.168.2.3");
server.bind(constants.daemon_port);

setInterval(heartbeat, 1000);
setInterval(dispatch_coldstart_request, 1000)