import sys
import json, ast
from RTEInterface import RTEInterface
from kafka import KafkaConsumer
import json
import time

f = open('../../constants_local.json')

constants = json.load(f)

consumer = KafkaConsumer('deployed3', 'removeWorker2', 'COLDSTART_WORKER2', 'AUTOSCALE2', constants['topics']['update_function_instance_nic'],
                         "request2", bootstrap_servers='10.129.2.201:9092')

RTEInterface.Connect('thrift', "10.129.2.201", 20206)
tableIdDispatch = "ingress::dispatch"
tableId2 = "ingress::fwd"
tableIdWarmstartDispatch = "ingress::warmstart_dispatch"

dispatchRuleDictionary = {}
warmstartDispatchRuleDictionary = {}
resourceMap = {}
functionToResourceMap = {}

def makeRuleForDispatchTable(ip, port, mac, functionHash, tableId, rule_name, default_rule):
    rule={}
    if default_rule:
        actions = '{  "type" : "ingress::dispatch_act",  "data" : {  "dstAddr" : { "value" : "%s" }, \
            "dstPort" : { "value" : "%d" } , "egress_port": { "value": "v0.1" }, "ethernetAddr": { "value": "%s" } } }' \
            % (ip, int(port), mac)
        match = '{ }'
        rule = {
            "tableId": tableId,
            "rule_name": rule_name,
            "default_rule": default_rule,
            "match": match,
            "actions": actions  
        }
    else:
        actions = '{  "type" : "ingress::dispatch_act",  "data" : {  "dstAddr" : { "value" : "%s" }, \
            "dstPort" : { "value" : "%d" } , "egress_port": { "value": "v0.1" }, "ethernetAddr": { "value": "%s" } } }' \
            % (ip, int(port), mac)
        match = '{ "map_hdr.function_id" : {  "value" : %d} }' % (functionHash)
        rule = {
            "tableId": tableId,
            "rule_name": rule_name,
            "default_rule": default_rule,
            "match": match,
            "actions": actions  
        }
    return rule

def addRuleForDispatchTable(worker):
    functionHash = worker[u'functionHash']
    rule_name = "dispatch_to_worker" + functionHash
    # functionHash = int(functionHash[0:5], 16)
    functionHash = int(functionHash[9:])
    ip = str(worker[u'node_id']).strip()
    port = int(worker[u'portExternal'])
    mac = str(worker[u'mac']).strip()
    resource_id = worker[u'resource_id']
    resourceMap[resource_id] = {
        "ip": ip,
        "port": port,
        "mac": mac
    }
    if(functionHash not in functionToResourceMap):
        functionToResourceMap[functionHash]= []
        
    functionToResourceMap[functionHash].append({"resource_id": resource_id})

    print("\n\nresource map : ", resourceMap)
    print("\n\nfunction to resource map : ", functionToResourceMap)

    default_rule = False
    rule = makeRuleForDispatchTable(ip, port, mac, functionHash, tableIdDispatch, rule_name, default_rule)
    dispatchRuleDictionary[str(functionHash)+resource_id] = rule
    print "\n\nrule added: ", dispatchRuleDictionary
    RTEInterface.Tables.AddRule(rule["tableId"], rule["rule_name"], rule["default_rule"], rule["match"], rule["actions"])
    ruleList = RTEInterface.Tables.ListRules(tableIdDispatch)
    print "\n\nnew rule list: ", ruleList, "\n\n"
    return 0


def deleteRuleFromDispatchTable(worker):
    functionHash = worker[u'functionHash']
    # functionHash = int(functionHash[0:5], 16)
    functionHash = int(functionHash[9:])
    resource_id = worker[u'resource_id']
    rule = dispatchRuleDictionary[str(functionHash)+resource_id]
    RTEInterface.Tables.DeleteRule(
        rule["tableId"], rule["rule_name"], rule["default_rule"], rule["match"], rule["actions"])
    del dispatchRuleDictionary[str(functionHash)+resource_id]

    del resourceMap[resource_id]
    for i in functionToResourceMap[functionHash]:
        if resource_id == i["resource_id"]:
            del i
    
    print("\n\nresource map : ", resourceMap)
    print("\n\nfunction to resource map : ", functionToResourceMap)

    ruleList = RTEInterface.Tables.ListRules(tableIdDispatch)
    print "\n\ndeleted rule"
    print "\n\nnew rule list: ", ruleList, "\n\n"
    
    return 0

def updateDefaultRuleOfDispatchTable(worker):
    # TODO 1: implement function to update deault rule when worker load heap changes
    # def EditRule(self, tbl_id, rule_name, default_rule, match, actions, priority = None, timeout = None):
    rule_name = "default"
    # functionHash = int(functionHash[0:5], 16)
    ip = str(worker[u'node_id']).strip()
    port = int(worker[u'portExternal'])
    mac = str(worker[u'mac']).strip()
    default_rule = True
    # autoscaling = 0

    rule = makeRuleForDispatchTable(ip, port, mac, None, tableIdDispatch, rule_name, default_rule)

    RTEInterface.Tables.EditRule(rule["tableId"], rule["rule_name"], rule["default_rule"], rule["match"], rule["actions"])
    print("\n\ndispatch table Default rule updated\n\n")
    return 0

def updateRuleOfDispatchTable(worker):
    # TODO 1: implement function to update deault rule when worker load heap changes
    # def EditRule(self, tbl_id, rule_name, default_rule, match, actions, priority = None, timeout = None):
    functionHash = worker[u'functionHash']
    rule_name = rule_name = "dispatch_to_worker" + functionHash
    # rule_name = 'dispatch_to_worker5a1'
    functionHash = int(functionHash[9:])
    
    ip = str(worker[u'node_id']).strip()
    port = int(worker[u'portExternal'])
    mac = str(worker[u'mac']).strip()
    # autoscaling = len(functionToResourceMap[functionHash])
    default_rule = False
    rule = makeRuleForDispatchTable(ip, port, mac, functionHash, tableIdDispatch, rule_name, default_rule)
    print("\n\nupdate rule : ", rule)
    RTEInterface.Tables.EditRule(rule["tableId"], rule["rule_name"], rule["default_rule"], rule["match"], rule["actions"])
    print("\n\dispatch table rule updated\n\n")
    return 0

def makeFwdTableRule(rule_name, default_rule):
    rule={}
    if default_rule:
        actions = '{  "type" : "ingress::fwd_act",  "data" : { "port" : { "value" : "p0" } } }'
        match = '{ "standard_metadata.ingress_port" : {  "value" : "v0.0" } }'
        rule = {
            "tableId": tableId2,
            "rule_name": 'host_to_net',
            "default_rule": default_rule,
            "match": match,
            "actions": actions  
        }
    else:
        actions = '{  "type" : "ingress::fwd_act",  "data" : { "port" : { "value" : "p0" } } }'
        match = '{ "standard_metadata.ingress_port" : {  "value" : "v0.0" } }'
        rule = {
            "tableId": tableId2,
            "rule_name": 'host_to_net',
            "default_rule": default_rule,
            "match": match ,
            "actions": actions  
        }
    return rule


for msg in consumer:
    if msg.topic == "deployed3":
        msg = msg.value.decode('utf-8')
        worker = json.loads(msg)
        print("received message on deployed : ", worker)
        # functionHash = worker[u'functionHash']
        # functionHash = int(functionHash[9:])
        # if(functionHash not in functionToResourceMap):
        addRuleForDispatchTable(worker)
            # addRuleForWarmstartDispatchTable(worker)
        # else:
            # addRuleForWarmstartDispatchTable(worker)
            # updateRuleOfDispatchTable(worker)
    
    elif msg.topic == "removeWorker2":
        msg = msg.value.decode('utf-8')
        worker = json.loads(msg)
        print("received message on removeWorker : ", worker)
        deleteRuleFromDispatchTable(worker)

    elif msg.topic == "COLDSTART_WORKER2":
        msg = msg.value.decode('utf-8')
        worker = json.loads(msg)
        print("received message on COLDSTART_WORKER : ", worker)
        updateDefaultRuleOfDispatchTable(worker[u'nodes'])

    elif msg.topic == constants['topics']['update_function_instance_nic']:
        msg = msg.value.decode('utf-8')
        worker = json.loads(msg)
        print("\nreceived message on UPDATE_FUNCTION_INSTANCE_NIC", worker)
        updateRuleOfDispatchTable(worker)

    
    


    # elif msg.topic == "AUTOSCALE":
        # msg = msg.value.decode('utf-8')
        
    
# updateDefaultRule2()
# worker1={
#     'node_id': '192.168.2.3',
#     'portExternal': '8081',
#     'mac': '00:22:22:22:22:22',
#     'resource_id': 'resource_id1',
#     'functionHash': 'function_1234'
# }

# worker2={
#     'node_id': '192.168.2.3',
#     'portExternal': '8082',
#     'mac': '00:22:22:22:22:22',
#     'resource_id': 'resource_id2',
#     'functionHash': 'function_1234'
# }
# while(1):
    
#     updateRuleOfDispatchTable(worker1)
#     time.sleep(2)
#     updateRuleOfDispatchTable(worker2)
#     time.sleep(2)

# addRuleForDispatchTable(worker1)
# addRuleForWarmstartDispatchTable(worker1)

# worker2={
#     'node_id': '192.168.2.3',
#     'portExternal': '8082',
#     'mac': '00:22:22:22:22:22',
#     'resource_id': 'resource_id2',
#     'functionHash': 'function_1234'
# }

# # addRuleForDispatchTable(worker2)
# addRuleForWarmstartDispatchTable(worker2)
# updateRuleOfDispatchTable(worker2)
f.close()